@@ -60,6 +60,8 @@ struct _TangleActorPrivate {
 	
 	GList* ancestor_events;
 	GList* descendant_events;
+	
+	ClutterAnimation* hiding_animation;
 
 	guint transition_scale_children : 1;
 	guint margin_set : 1;
@@ -68,6 +70,8 @@ struct _TangleActorPrivate {
 	guint interacting : 1;
 	guint interactive_size_multiplier_set : 1;
 	guint propagating_events : 1;
+	guint hide_all : 1;
+	guint destroy : 1;
 };
 
 enum {
@@ -116,6 +120,7 @@ static void set_picking_scale_center_by_gravity(TangleActor* actor);
 static void set_picking_scale_gravity(TangleActor* actor, ClutterGravity gravity);
 static gboolean handle_queued_events(TangleActor* actor);
 static void unset_transition_scale_children(TangleActor* actor);
+static void on_hiding_animation_completed(TangleActor* actor);
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
@@ -175,23 +180,55 @@ gboolean tangle_actor_should_transition_scale_children(TangleActor* actor) {
 	return actor->priv->transition_scale_children;
 }
 
+void tangle_actor_show(TangleActor* actor) {
+	ClutterActorBox* current_allocation;
+	ClutterAnimation* animation = NULL;
+
+	if (TANGLE_ACTOR_IS_HIDING(actor) && !actor->priv->destroy) {
+		TANGLE_ACTOR_UNSET_FLAGS(actor, TANGLE_ACTOR_HIDING);
+
+		clutter_animation_completed(actor->priv->hiding_animation);
+		g_object_unref(actor->priv->hiding_animation);
+		actor->priv->hiding_animation = NULL;
+
+		g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
+		actor->priv->transition_scale_children = TRUE;
+		g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, NULL, current_allocation, &animation);
+		if (animation) {
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
+		}
+		
+		/* TODO: if hiding_animation should have done hide_all... */
+	}
+
+	if (!actor->priv->destroy && !CLUTTER_ACTOR_IS_VISIBLE(CLUTTER_ACTOR(actor))) {
+		clutter_actor_show(CLUTTER_ACTOR(actor));
+	}
+}
+
 void tangle_actor_hide_animated(TangleActor* actor) {
 	ClutterActorBox* current_allocation;
 	ClutterAnimation* animation = NULL;
 	
 	g_return_if_fail(TANGLE_IS_ACTOR(actor));
 	
-	if (TANGLE_ACTOR_IS_PAINTED(actor)) {
-		g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
-		actor->priv->transition_scale_children = TRUE;
-		g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, current_allocation, NULL, &animation);
-	}
-	if (animation) {
-		g_signal_connect_swapped(animation, "completed", G_CALLBACK(clutter_actor_hide), actor);
-		g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
-	} else {
-		clutter_actor_hide(CLUTTER_ACTOR(actor));
-		actor->priv->transition_scale_children = FALSE;
+	if (CLUTTER_ACTOR_IS_VISIBLE(CLUTTER_ACTOR(actor)) && !TANGLE_ACTOR_IS_HIDING(actor)) {
+		if (TANGLE_ACTOR_IS_PAINTED(actor)) {
+			g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
+			actor->priv->transition_scale_children = TRUE;
+			g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, current_allocation, NULL, &animation);
+		}
+		if (animation) {
+			actor->priv->hiding_animation = animation;
+			g_object_ref(animation);
+			TANGLE_ACTOR_SET_FLAGS(actor, TANGLE_ACTOR_HIDING);
+
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(on_hiding_animation_completed), actor);
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
+		} else {
+			clutter_actor_hide(CLUTTER_ACTOR(actor));
+			actor->priv->transition_scale_children = FALSE;
+		}
 	}
 }
 
@@ -199,17 +236,27 @@ void tangle_actor_hide_all_animated(TangleActor* actor) {
 	ClutterActorBox* current_allocation;
 	ClutterAnimation* animation = NULL;
 	
-	if (TANGLE_ACTOR_IS_PAINTED(actor)) {
-		g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
-		actor->priv->transition_scale_children = TRUE;
-		g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, current_allocation, NULL, &animation);
-	}
-	if (animation) {
-		g_signal_connect_swapped(animation, "completed", G_CALLBACK(clutter_actor_hide_all), actor);
-		g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
-	} else {
+	actor->priv->hide_all = TRUE;
+	
+	if (CLUTTER_ACTOR_IS_VISIBLE(CLUTTER_ACTOR(actor)) && !TANGLE_ACTOR_IS_HIDING(actor)) {
+		if (TANGLE_ACTOR_IS_PAINTED(actor)) {
+			g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
+			actor->priv->transition_scale_children = TRUE;
+			g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, current_allocation, NULL, &animation);
+		}
+		if (animation) {
+			actor->priv->hiding_animation = animation;
+			g_object_ref(animation);
+			TANGLE_ACTOR_SET_FLAGS(actor, TANGLE_ACTOR_HIDING);
+
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(on_hiding_animation_completed), actor);
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
+		} else {
+			clutter_actor_hide_all(CLUTTER_ACTOR(actor));
+			actor->priv->transition_scale_children = FALSE;
+		}
+	} else if (!CLUTTER_ACTOR_IS_VISIBLE(CLUTTER_ACTOR(actor))) {
 		clutter_actor_hide_all(CLUTTER_ACTOR(actor));
-		actor->priv->transition_scale_children = FALSE;
 	}
 }
 
@@ -217,17 +264,27 @@ void tangle_actor_destroy_animated(TangleActor* actor) {
 	ClutterActorBox* current_allocation;
 	ClutterAnimation* animation = NULL;
 	
-	if (TANGLE_ACTOR_IS_PAINTED(actor)) {
-		g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
-		actor->priv->transition_scale_children = TRUE;
-		g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, current_allocation, NULL, &animation);
-	}
-	if (animation) {
-		g_signal_connect_swapped(animation, "completed", G_CALLBACK(clutter_actor_destroy), actor);
-		g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
-	} else {
+	actor->priv->destroy = TRUE;
+	
+	if (CLUTTER_ACTOR_IS_VISIBLE(CLUTTER_ACTOR(actor)) && !TANGLE_ACTOR_IS_HIDING(actor)) {
+		if (TANGLE_ACTOR_IS_PAINTED(actor)) {
+			g_object_get(G_OBJECT(actor), "allocation", &current_allocation, NULL);
+			actor->priv->transition_scale_children = TRUE;
+			g_signal_emit(actor, signals[ANIMATE_TRANSITION], 0, current_allocation, NULL, &animation);
+		}
+		if (animation) {
+			actor->priv->hiding_animation = animation;
+			g_object_ref(animation);
+			TANGLE_ACTOR_SET_FLAGS(actor, TANGLE_ACTOR_HIDING);
+
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(on_hiding_animation_completed), actor);
+			g_signal_connect_swapped(animation, "completed", G_CALLBACK(unset_transition_scale_children), actor);
+		} else {
+			clutter_actor_destroy(CLUTTER_ACTOR(actor));
+			actor->priv->transition_scale_children = FALSE;
+		}
+	} else if (!CLUTTER_ACTOR_IS_VISIBLE(CLUTTER_ACTOR(actor))) {
 		clutter_actor_destroy(CLUTTER_ACTOR(actor));
-		actor->priv->transition_scale_children = FALSE;
 	}
 }
 
@@ -595,6 +652,7 @@ static void on_animation_completed(ClutterAnimation* animation, gpointer user_da
 	TangleActor* actor;
 
 	actor = TANGLE_ACTOR(user_data);
+
 	g_hash_table_foreach_remove(actor->priv->animations, is_value_user_data, animation);
 	g_object_unref(animation);
 }
@@ -739,11 +797,27 @@ static void tangle_actor_paint(ClutterActor* clutter_actor) {
 	actor_class->paint_aligned(actor, width, height);
 }
 
+static void tangle_actor_pick(ClutterActor* clutter_actor, const ClutterColor* color) {
+	if (!TANGLE_ACTOR_IS_HIDING(TANGLE_ACTOR(clutter_actor))) {
+		CLUTTER_ACTOR_CLASS(tangle_actor_parent_class)->pick(clutter_actor, color);
+	}
+}
+
 static void tangle_actor_hide(ClutterActor* clutter_actor) {
 	TangleActor* actor;
 	
 	actor = TANGLE_ACTOR(clutter_actor);
+
 	TANGLE_ACTOR_UNSET_FLAGS(actor, TANGLE_ACTOR_PAINTED);
+
+	if (TANGLE_ACTOR_IS_HIDING(actor)) {
+		TANGLE_ACTOR_UNSET_FLAGS(actor, TANGLE_ACTOR_HIDING);
+
+		/* Unref also animation's internal reference to stop it. */
+		g_object_unref(actor->priv->hiding_animation);
+		g_object_unref(actor->priv->hiding_animation);
+		actor->priv->hiding_animation = NULL;
+	}
 	
 	CLUTTER_ACTOR_CLASS(tangle_actor_parent_class)->hide(clutter_actor);
 }
@@ -1098,6 +1172,7 @@ static void tangle_actor_class_init(TangleActorClass* actor_class) {
 	clutter_actor_class->get_preferred_height = tangle_actor_override_get_preferred_height;
 	clutter_actor_class->allocate = tangle_actor_allocate;
 	clutter_actor_class->paint = tangle_actor_paint;
+	clutter_actor_class->pick = tangle_actor_pick;
 	clutter_actor_class->hide = tangle_actor_hide;
 	clutter_actor_class->button_press_event = tangle_actor_button_press_event;
 	clutter_actor_class->button_release_event = tangle_actor_button_press_event;
@@ -1411,3 +1486,21 @@ static gboolean handle_queued_events(TangleActor* actor) {
 static void unset_transition_scale_children(TangleActor* actor) {
 	actor->priv->transition_scale_children = FALSE;
 }
+
+static void on_hiding_animation_completed(TangleActor* actor) {
+	if (TANGLE_ACTOR_IS_HIDING(actor)) {
+		TANGLE_ACTOR_UNSET_FLAGS(actor, TANGLE_ACTOR_HIDING);
+
+		g_object_unref(actor->priv->hiding_animation);
+		actor->priv->hiding_animation = NULL;
+
+		if (actor->priv->destroy) {
+			clutter_actor_destroy(CLUTTER_ACTOR(actor));
+		} else if (actor->priv->hide_all) {
+			clutter_actor_hide_all(CLUTTER_ACTOR(actor));
+		} else {
+			clutter_actor_hide(CLUTTER_ACTOR(actor));
+		}
+	}
+	actor->priv->hide_all = FALSE;
+}
